import { Component, OnInit } from "@angular/core";

import { ParamDomainValue, Observer } from "jalhyd";

import { ApplicationSetupService } from "../../services/app-setup/app-setup.service";
import { I18nService } from "../../services/internationalisation/internationalisation.service";
import { NgBaseParam } from "../base-param-input/base-param-input.component";
import { BaseComponent } from "../base/base.component";
import { ErrorStateMatcher, MatSnackBar } from "@angular/material";


@Component({
    selector: "setup",
    templateUrl: "./app-setup.component.html",
    styleUrls: ["./app-setup.component.scss"]
})
export class ApplicationSetupComponent extends BaseComponent implements Observer, OnInit {

    /** précision d'affichage */
    public displayPrec: NgBaseParam;

    /** précision de calcul */
    public computePrec: NgBaseParam;

    /** nombre d'iterations max Newton */
    public newtonMaxIter: NgBaseParam;

    public matcher: ErrorStateMatcher;

    constructor(
        private appSetupService: ApplicationSetupService,
        private intlService: I18nService,
        private snackBar: MatSnackBar
    ) {
        super();
        this.appSetupService.addObserver(this);
    }

    public get availableLanguages() {
        return this.intlService.languages;
    }

    public get currentLanguageCode() {
        if (this.intlService.currentLanguage) {
            return this.intlService.currentLanguage;
        }
    }

    public set currentLanguageCode(lc: string) {
        this.intlService.setLanguage(lc);
        // keep language in sync in app-wide parameters service
        this.appSetupService.language = this.intlService.currentLanguage;
    }

    public get uitextTitle(): string {
        return this.intlService.localizeText("INFO_SETUP_TITLE");
    }

    public get uitextDisplayAccuracy(): string {
        return this.intlService.localizeText("INFO_SETUP_PRECISION_AFFICHAGE");
    }

    public get uitextComputeAccuracy(): string {
        return this.intlService.localizeText("INFO_SETUP_PRECISION_CALCUL");
    }

    public get uitextNewtonMaxIteration(): string {
        return this.intlService.localizeText("INFO_SETUP_NEWTON_MAX_ITER");
    }

    public get uitextRememberValues(): string {
        return this.intlService.localizeText("INFO_MENU_SAVE_SETTINGS");
    }

    public get uitextRestoreDefault(): string {
        return this.intlService.localizeText("INFO_MENU_RESTORE_DEFAULT_SETTINGS");
    }

    public get uitextMustBeANumber(): string {
        return this.intlService.localizeText("ERROR_PARAM_MUST_BE_A_NUMBER");
    }

    public storePreferences() {
        this.appSetupService.saveValuesIntoLocalStorage();
        this.snackBar.open(this.intlService.localizeText("INFO_SNACKBAR_SETTINGS_SAVED"), "OK", {
            duration: 2500
        });
    }

    public restoreDefaultValues() {
        const text = this.intlService.localizeText("INFO_SNACKBAR_DEFAULT_SETTINGS_RESTORED");
        this.appSetupService.restoreDefaultValues().then(() => {
            this.snackBar.open(text, "OK", {
                duration: 2500
            });
        });
    }

    private init() {
        // modèle du composant BaseParamInputComponent de précision d'affichage
        this.displayPrec = new NgBaseParam("dp", ParamDomainValue.POS, this.appSetupService.displayPrecision);
        this.displayPrec.addObserver(this);

        // modèle du composant BaseParamInputComponent de précision de calcul
        this.computePrec = new NgBaseParam("cp", ParamDomainValue.POS, this.appSetupService.computePrecision);
        this.computePrec.addObserver(this);

        // modèle du composant BaseParamInputComponent du max d'itérations pour Newton
        this.newtonMaxIter = new NgBaseParam("nmi", ParamDomainValue.POS, this.appSetupService.newtonMaxIterations);
        this.newtonMaxIter.addObserver(this);
    }

    ngOnInit() {
        this.init();
    }

    // interface Observer
    public update(sender: any, data: any): void {
        if (sender instanceof NgBaseParam) {
            // someone typed a new value in the app-setup form
            const p: NgBaseParam = sender;
            switch (p.symbol) {
                case "dp":
                    this.appSetupService.displayPrecision = +data;
                    break;

                case "cp":
                    this.appSetupService.computePrecision = +data;
                    break;

                case "nmi":
                    this.appSetupService.newtonMaxIterations = +data;
                    break;
            }
        }
        if (sender instanceof ApplicationSetupService) {
            // if app starts on /setup page, wait for config file to be (re)loaded by app-setup service
            this.init();
        }
    }
}
