import { Nub, Result, ParamDomainValue, Observer, ParamDefinition } from "jalhyd";

import { FormResult } from "./form-result";
import { FormulaireDefinition } from "./form-definition";

export abstract class FormCompute implements Observer {

    constructor(protected _formBase: FormulaireDefinition, protected _formResult: FormResult) {
        // indirectly subscribe to Nub result updates
        this._formBase.addObserver(this);
    }

    protected abstract compute();

    protected get formResult(): FormResult {
        return this._formResult;
    }

    /**
     * retourne un identifiant du paramètre dans le formulaire
     * surchargé dans le cas des ouvrages //
     */
    protected getParameterRefid(p: ParamDefinition) {
        return p.symbol;
    }

    /**
     * Copies current Nub result into result components for display on page.
     * Should be called every time the Nub result changes.
     * Must be idempotent.
     */
    protected abstract reaffectResultComponents();

    /**
     * Lance le calcul d'un paramètre en déterminant une valeur initiale.
     * Si nécessaire déclenche un calcul en chaîne des modules en amont.
     */
    protected runNubCalc(nub: Nub, computedParam?: ParamDefinition): Result {
        let init: number;

        // by default, use Nub's calculatedParam
        if (computedParam === undefined) {
            computedParam = nub.calculatedParam;
        }

        // require chain computation; redundant with Nub.CalcSerie but required
        // to get initial value here...
        const computedParamValue = computedParam.getValue();

        switch (computedParam.domain.domain) {
            case ParamDomainValue.ANY:
                if (computedParam && computedParam.isDefined) {
                    init = computedParamValue;
                }
                if (init === undefined) {
                    init = 0;
                }
                break;

            case ParamDomainValue.POS_NULL:
                if (computedParam && computedParam.isDefined) {
                    init = Math.max(computedParamValue, 0);
                }
                if (init === undefined) {
                    init = 0;
                }
                break;

            case ParamDomainValue.INTERVAL:
                init = (computedParam.domain.minValue + computedParam.domain.maxValue) / 2;
                break;

            case ParamDomainValue.NOT_NULL:
                if (computedParam && computedParam.isDefined) {
                    init = computedParamValue;
                }
                if (init === undefined || init === 0) {
                    init = 1e-8;
                }
                break;

            case ParamDomainValue.POS:
                if (computedParam && computedParam.isDefined) {
                    init = Math.max(computedParamValue, 1e-8);
                }
                if (init === undefined) {
                    init = 1e-8;
                }
                break;
        }

        return nub.CalcSerie(init, this.getParameterRefid(computedParam));
    }

    /**
     * Triggers computation of the Nub, updates form results
     */
    public doCompute() {
        this._formResult.resetResults();

        this.compute();

        this._formBase.notifyObservers({
            "action": "resultsUpdated",
        }, this._formBase);
    }

    // interface Observer

    public update(sender: any, data: any): void {
        if (sender instanceof Nub) {
            switch (data.action) {
                case "nubResultUpdated":
                    // forward Nub results update notification to FormCompute objects
                    this.reaffectResultComponents();
                    break;
            }
        }
    }
}
